/* Standard includes... */
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* EPICS includes */
#include <epicsString.h>
#include <epicsTime.h>
#include <epicsThread.h>
#include <epicsEvent.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <epicsExit.h>
#include <envDefs.h>


/* Area Detector includes */
#include <ADStdDriverParams.h>
#include <NDArray.h>
#include <ADDriver.h>

#include "adXmap.h"

/** Only used for debugging/error messages to identify where the message comes from*/
static const char *driverName = "ADXmap";
static ENV_PARAM XIAHOME;

typedef enum ADXmode_t { ADX_normal = 0, ADX_mapping }ADXmode_t;

typedef enum ADXParam_t {
	ADX_xmap_mode = ADFirstDriverParam, /** < Change mode of the XMAP (0=mca; 1=mapping; 2=sca) (int32 read/write) addr: all/any */
	ADX_start_mode,                     /** < Start mode. (0=clear on start; 1=resume spectrum) */
	ADX_xmap_runstate,                  /** < XMAP reporting it's runtime state (int bitmap) */
	ADX_current_pixel,                  /** < XMAP mapping mode only: read the current pixel that is being acquired into (int) */
	ADX_next_pixel,                     /** < XMAP mapping mode only: force a pixel increment in the xmap buffer (write only int). Value is ignored. */
	ADX_buffer_overrun,
	ADX_mbytes_received,
	ADX_read_speed,

	/* runtime statistics */
	ADX_stat_realtime,                  /** < real time in seconds (double) */
	ADX_stat_triggerlivetime,           /** < live time in seconds (double) */
	ADX_stat_energylivetime,            /** < live time in seconds (double) */
	ADX_stat_triggers,                  /** < number of triggers received (int) */
	ADX_stat_events,                    /** < total number of events registered (int) */
	ADX_stat_inputcountrate,            /** < input count rate in Hz (double) */
	ADX_stat_outputcountrate,           /** < output count rate in Hz (double) */

	ADX_xia_peaking_time,
	ADX_xia_dynamic_range,
	ADX_xia_trigger_threshold,
	ADX_xia_baseline_threshold,
	ADX_xia_energy_threshold,
	ADX_xia_calibration_energy,
	ADX_xia_adc_percent_rule,
	ADX_xia_mca_bin_width,
	ADX_xia_preamp_gain,
	ADX_xia_number_mca_channels,
	ADX_xia_detector_polarity,
	ADX_xia_reset_delay,
	ADX_xia_gap_time,
	ADX_xia_trigger_peaking_time,
	ADX_xia_trigger_gap_time,
	ADX_xia_baseline_average,

	ADLastDriverParam
	} ADXParam_t;

static asynParamString_t ADXParamString[] = {
	{ADX_xmap_mode,                "ADX_XMAP_MODE"},
	{ADX_start_mode,               "ADX_START_MODE"},
	{ADX_xmap_runstate,            "ADX_XMAP_RUNSTATE"},
	{ADX_current_pixel,            "ADX_CURRENT_PIXEL"},
	{ADX_next_pixel,               "ADX_NEXT_PIXEL"},
	{ADX_buffer_overrun,           "ADX_BUFFER_OVERRUN"},
	{ADX_mbytes_received,          "ADX_MBYTES_RECEIVED"},
	{ADX_read_speed,               "ADX_READ_SPEED"},

	{ADX_stat_realtime,            "ADX_STAT_REALTIME"},
	{ADX_stat_triggerlivetime,     "ADX_STAT_TRIGLTIME"},
	{ADX_stat_energylivetime,      "ADX_STAT_ELTIME"},
	{ADX_stat_triggers,            "ADX_STAT_TRIGGERS"},
	{ADX_stat_events,              "ADX_STAT_EVENTS"},
	{ADX_stat_inputcountrate,      "ADX_STAT_INCNTRATE"},
	{ADX_stat_outputcountrate,     "ADX_STAT_OUTCNTRATE"},

	{ADX_xia_peaking_time,         "peaking_time"},
	{ADX_xia_dynamic_range,        "dynamic_range"},
	{ADX_xia_trigger_threshold,    "trigger_threshold"},
	{ADX_xia_baseline_threshold,   "baseline_threshold"},
	{ADX_xia_energy_threshold,     "energy_threshold"},
	{ADX_xia_calibration_energy,   "calibration_energy"},
	{ADX_xia_adc_percent_rule,     "adc_percent_rule"},
	{ADX_xia_mca_bin_width,        "mca_bin_width"},
	{ADX_xia_preamp_gain,          "preamp_gain"},
	{ADX_xia_number_mca_channels,  "number_mca_channels"},
	{ADX_xia_detector_polarity,    "detector_polarity"},
	{ADX_xia_reset_delay,          "reset_delay"},
	{ADX_xia_gap_time,             "gap_time"},
	{ADX_xia_trigger_peaking_time, "trigger_peaking_time"},
	{ADX_xia_trigger_gap_time,     "trigger_gap_time"},
	{ADX_xia_baseline_average,     "baseline_average"},
	};

typedef struct ADXBuffers_t {
	char bufChar[2];
	char bufFull[14];
	char buffer[9];
} ADXBuffers_t;

static ADXBuffers_t ADXBuffers[] =
{
		{"a", "buffer_full_a", "buffer_a"},
		{"b", "buffer_full_b", "buffer_b"},
};

/** Number of asyn parameters (asyn commands) this driver supports. */
#define ADX_N_PARAMS (sizeof( ADXParamString)/ sizeof(ADXParamString[0]))


class ADXmap : public ADDriver
{
public:
	ADXmap(const char *portName, int nCards, const char *iniFile, int maxBuffers, size_t maxMemory);

	/* virtual methods to override from ADDriver */
	virtual asynStatus writeInt32( asynUser *pasynUser, epicsInt32 value);
	virtual asynStatus writeFloat64( asynUser *pasynUser, epicsFloat64 value);
	virtual asynStatus drvUserCreate( asynUser *pasynUser, const char *drvInfo, const char **pptypeName, size_t *psize);
	void report(FILE *fp, int details);

	/* Local methods to this class */
	asynStatus err( asynUser* asynUser, int errOriginLine);
	asynStatus xia_checkError( asynUser* pasynUser, epicsInt32 xiastatus, char *xiacmd );
	void shutdown();

	void acquisitionTask();
	asynStatus pollMcaMode();
	asynStatus pollMappingMode();
	asynStatus getStatus(asynUser* pasynUser);

	asynStatus changeMode(asynUser *pasynUser, epicsInt32 mode);
	asynStatus calcMaxY(asynUser *pasynUser);

	/* Get the status of XMAP both at runtime and when idle. Intended to be called in a
	 * polling fashion from the acquisition thread. */
	asynStatus statusUpdate(asynUser *pasynUser);

	asynStatus changeSizeX(asynUser *pasynUser, epicsInt32 newsize, epicsInt32 *rbValue);

private:
	/* Data */
	NDArray *pRaw;
	epicsUInt32 *pMcaRaw;
	epicsUInt32 *pMapRaw;
	epicsUInt16 *pMapSmallWord;
	epicsFloat64 *tmpStats;

	epicsUInt32 nCards;
	epicsUInt32 nChannels;
	char * iniFile;

    epicsEventId startEventId;
    epicsEventId stopEventId;

    epicsMessageQueue *bufReadQueue;
    epicsUInt32 msgMaxSize;

    epicsUInt32 *currentBuf;
    epicsTimeStamp tStartAcq;
};

static void c_shutdown(void* arg)
{
	ADXmap *adXmap = (ADXmap*)arg;
	adXmap->shutdown();
}

static void acquisitionTaskC(void *drvPvt)
{
    ADXmap *adXmap = (ADXmap *)drvPvt;
    adXmap->acquisitionTask();
}

extern "C" int ADXmap_config(const char *portName, int nCards, const char *iniFile, int maxBuffers, size_t maxMemory)
{
	new ADXmap(portName, nCards, iniFile, maxBuffers, maxMemory);
	return 0;
}

ADXmap::ADXmap(const char *portName, int nCards, const char *iniFile, int maxBuffers, size_t maxMemory)
	: ADDriver(portName, nCards * XMAP_NCHANS_MODULE, ADLastDriverParam, maxBuffers, maxMemory, 0, 0),
	pRaw(NULL)
{
	int status = asynSuccess;
	int xiastatus = 0;
	static int configured = FALSE;
	char * xia_home_path;
	epicsTimeStamp startTime;
	char *strStartTime, *logFileName;
	int dimensions[2];
	const char *functionName = "ADXmap";

	this->nCards = nCards;
	this->nChannels = this->nCards * XMAP_NCHANS_MODULE;
	this->iniFile = epicsStrDup(iniFile);


	/* if it is the first invocation of the config function,
	 * we initialize the xmap/handel library
	 * - and create the linked list of driver structures. */
	if (configured == FALSE)
	{
		printf("  Initializing Handel with \"%s\"\n", this->iniFile);
		xiastatus = xiaInit(this->iniFile);
		status = this->xia_checkError( this->pasynUserSelf, xiastatus, "xiaInit()");
		if (status == asynError) return; /* Giving up.... won't work!!! */

		/* Register the epics exit function to be called when the IOC exits... */
		xiastatus = epicsAtExit(c_shutdown, (void*)this);
		printf("  epicsAtExit registered: %d\n", xiastatus);

		xiastatus = xiaSetLogLevel(MD_ERROR);
		status = this->xia_checkError( this->pasynUserSelf, xiastatus, "xiaSetLogLevel()");

		/* Get the XIAHOME environment variable to define where to write the
		 * error log files to.		 */
		XIAHOME.name = epicsStrDup("XIAHOME");
		XIAHOME.pdflt = (char*)calloc(256, sizeof(char));
		xia_home_path = (char*)calloc(256, sizeof(char));
		envGetConfigParam(&XIAHOME, 256, xia_home_path);
		epicsTimeGetCurrent(&startTime);
		strStartTime = (char*)calloc(80, sizeof(char));
		epicsTimeToStrftime ( strStartTime, 80, "%Y-%m-%d_%H-%M-%S", &startTime );
		logFileName = (char*)calloc(256, sizeof(char));
		sprintf(logFileName, "%s\\epics-xmap_%s.log", xia_home_path, strStartTime);
		printf("  XIA log file: %s\n", logFileName);

		xiastatus = xiaSetLogOutput(logFileName);
		status = this->xia_checkError( this->pasynUserSelf, xiastatus, "xiaSetLogOutput()");

		printf("  Starting XIA system.                      ");
		xiastatus = xiaStartSystem();
		status = this->xia_checkError( this->pasynUserSelf, xiastatus, "xiaStartSystem()");
		if (status == asynError) return; /* Giving up - won't work!!! */
		printf("OK\n");

		configured = TRUE;
	}

	/* The default settings at wake up is in normal MCA mode.
	 * Which implies the Y dimension is just 1.	 */
	dimensions[0] = XMAP_MAX_MCA_BINS;
	dimensions[1] = 1;    /* TODO: Get the maximum number of spectrums in one 2MB buffer from the manual */

	/* Set the parameters from the camera in our areaDetector param lib */
	printf("  Setting the areaDetector parameters...   ");
    status =  setStringParam (ADManufacturer, "XIA");
    status |= setStringParam (ADModel, "XMAP");
    status |= setIntegerParam(ADMaxSizeX, dimensions[0]);
    status |= setIntegerParam(ADMaxSizeY, dimensions[1]);
    status |= setIntegerParam(ADSizeX, dimensions[0]);
    status |= setIntegerParam(ADSizeY, dimensions[1]);
    status |= setIntegerParam(ADImageSizeX, dimensions[0]);
    status |= setIntegerParam(ADImageSizeY, dimensions[1]);
    status |= setIntegerParam(ADImageSize, dimensions[0]*dimensions[1]*sizeof(char));
    status |= setIntegerParam(ADDataType, NDUInt32);
    status |= setIntegerParam(ADImageMode, ADImageContinuous);
    status |= setIntegerParam(ADNumImages, 100);

    status |= setIntegerParam(ADX_start_mode, 0);
    status |= setIntegerParam(ADX_xmap_mode, 0);
    printf("OK\n");

	/* Create the start and stop event that will be used to signal our
	 * image grabbing thread when to start/stop	 */
	printf("  Creating epicsevents...                  ");
	this->startEventId = epicsEventCreate(epicsEventEmpty);
	this->stopEventId = epicsEventCreate(epicsEventFull);
	printf("OK\n");

	printf("  Creating message queue...");
	this->msgMaxSize = sizeof(epicsInt32);
	this->bufReadQueue = new epicsMessageQueue(this->msgMaxSize, XMAP_NCHANS_MODULE);
	printf("OK\n");

	/* allocate memory for one board full MCA spectrum */
    this->pMcaRaw = (epicsUInt32*)calloc(XMAP_NCHANS_MODULE * XMAP_MAX_MCA_BINS, sizeof(epicsUInt32));
    this->tmpStats = (epicsFloat64*)calloc(28, sizeof(epicsFloat64));
    this->currentBuf = (epicsUInt32*)calloc(this->nChannels, sizeof(epicsUInt32));
    /* Allocating a temporary buffer for use when collecting mapped MCA spectrums.
     * The XMAP buffer takes up 2MB of 16bit words. Unfortunatly the transfer over PCI
     * uses 32bit words so the data we receive from from the Handel library is 2x2MB large.
     * Thus the 2MB buffer: XMAP_MAX_MAPBUF_SIZE = 2 * 1024 * 1024; */
    this->pMapRaw = (epicsUInt32*)malloc(2*XMAP_MAX_MAPBUF_SIZE);
    this->pMapSmallWord = (epicsUInt16*)calloc(XMAP_MAX_MAPBUF_SIZE, sizeof(epicsUInt16));
    printf("  Allocated mapping memory (4MB): %p\n", this->pMapRaw);

	/* Start up acquisition thread */
    printf("  Starting up acquisition task...          ");
    status = (epicsThreadCreate("acquisitionTask",
    		epicsThreadPriorityMedium,
    		epicsThreadGetStackSize(epicsThreadStackMedium),
    		(EPICSTHREADFUNC)acquisitionTaskC, this) == NULL);
    if (status)
    {
    	printf("%s:%s epicsThreadCreate failure for image task\n",
    			driverName, functionName);
    	return;
    } else printf("OK\n");

}

/* virtual methods to override from ADDriver */
asynStatus ADXmap::writeInt32( asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	int function = pasynUser->reason;
	int adstatus, eventStatus;
	int addr, rbValue, xiastatus;

	const char* functionName = "writeInt32";

	pasynManager->getAddr(pasynUser, &addr);

	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s [%s]: function= %d value=%d\n",
				driverName, functionName, this->portName, function, value);

	rbValue = value;

	switch(function)
	{
	case ADX_xmap_mode:
		status = this->changeMode(pasynUser, value);
		break;

	case ADX_start_mode:
		if (value == 1 || value == 0) {status = asynSuccess;}
		else {status = asynError;}
		break;

	case ADX_next_pixel:
		unsigned int firstCh, ignored;

		for (firstCh = 0; firstCh < this->nChannels; firstCh += XMAP_NCHANS_MODULE)
		{
			xiastatus = xiaBoardOperation(firstCh, "mapping_pixel_next", (void*)&ignored);
			status = this->xia_checkError(pasynUser, xiastatus, "mapping_pixel_next");
		}
		rbValue = 0;
		break;


	case ADAcquire:
		getIntegerParam(ADStatus, &adstatus);
		if (value && (adstatus == ADStatusIdle))
		{
			/* start acquisition */
			epicsEventSignal(this->startEventId);
		} else if (!value)
		{
			setIntegerParam(ADAcquire, 0);		/* set acquiring state to stopped */
			callParamCallbacks(); 				/* update whoever is interested in the acquire-state */

			/* Now wait for the capture thread to actually stop */
			asynPrint( pasynUser, ASYN_TRACE_ERROR, "%s::%s [%s] waiting for stopped event...\n",
							driverName, functionName, this->portName);
			/* unlock the mutex while we're waiting for the capture thread to stop acquiring */
			epicsMutexUnlock(this->mutexId);
			eventStatus = epicsEventWaitWithTimeout(this->stopEventId, 3.0);
			epicsMutexLock(this->mutexId);
			if (eventStatus != epicsEventWaitOK)
			{
				asynPrint( pasynUser, ASYN_TRACE_FLOW, "%s::%s [%s] ERROR: Timeout when trying to stop image grabbing thread.\n",
							driverName, functionName, this->portName);
			}
			/* Get the status again just after stopping */
			this->getStatus(pasynUser);
		}
		break;

	case ADSizeX:
		status = this->changeSizeX(pasynUser, value, &rbValue);
		if (status == asynSuccess)
		{
			setIntegerParam(ADImageSizeX, rbValue);
		}
		break;

	case ADSizeY:
		status = asynError; /* TODO: need to support this in mapping mode. */
		break;
	case ADImageMode:
	case ADNumImages:
	case ADAcquirePeriod:
	case ADNumImagesCounter:
	case ADTriggerMode:
	case ADArrayCallbacks:
		break;
	default:
		asynPrint(pasynUser, ASYN_TRACE_ERROR, "%s::%s Function not implemented: %d; val=%d\n",
					driverName, functionName, function, value);
		status = asynError;
		break;
	}

	//if (status != asynError) status = setIntegerParam(function, value);
	if (status != asynError) status = setIntegerParam(addr, function, rbValue);

	/* Call the callback for the specific address .. and address ... weird? */
	if (status != asynError) callParamCallbacks(addr, addr);
	return status;
}

asynStatus ADXmap::writeFloat64( asynUser *pasynUser, epicsFloat64 value)
{
	asynStatus status = asynSuccess;
	int function = pasynUser->reason;
	int addr;
	epicsFloat64 rbValue;
	const char* functionName = "writeFloat64";

	pasynManager->getAddr(pasynUser, &addr);
	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s [%s]: function=%d value=%.5f\n",
				driverName, functionName, this->portName, function, value);

	rbValue = value;

	switch(function)
	{

		case ADAcquirePeriod:
			if (value < 0.01 || value > 1)
			{
				asynPrint(pasynUser, ASYN_TRACE_ERROR, "ERROR: Value %d is out of range [0.01s .. 1s]\n");
				status = asynError;
			}
			break;
		default:
			asynPrint(pasynUser, ASYN_TRACE_ERROR, "%s::%s Function not implemented: %d; val=%.5f\n",
						driverName, functionName, function, value);
			status = asynError;
			break;
		}

	//if (status != asynError) status = setIntegerParam(function, value);
	if (status != asynError) status = setDoubleParam(addr, function, rbValue);

	/* Call the callback for the specific address .. and address ... weird? */
	if (status != asynError) callParamCallbacks(addr, addr);

	return status;
}


asynStatus ADXmap::changeSizeX(asynUser* pasynUser, epicsInt32 value, epicsInt32 *rbValue)
{
	asynStatus status = asynSuccess;
	int xiastatus;
	unsigned int i, bufLen;
	int modulus, realnbins, dummy, maxSizeY=0, mode=0;
	double dblNbins, dNumPixPerBuf = -1.0, dTmp;
	const char* functionName = "changeSizeX";

	if (value > XMAP_MAX_MCA_BINS || value < XMAP_MCA_BIN_RES)
	{
		status = asynError;
		return status;
	}

	/* TODO: check here whether the xmap is acquiring (if so, break out...)  */

	/* MCA number of bins on the xmap has a resolution of XMAP_MCA_BIN_RES (256)
	 * bins per channel. Since the user can enter any value we must cap this to
	 * the closest multiple of 256 */
	modulus = value % XMAP_MCA_BIN_RES;
	if (modulus == 0) realnbins = value;
	else if (modulus < XMAP_MCA_BIN_RES/2) realnbins = value - modulus;
	else realnbins = value + (XMAP_MCA_BIN_RES - modulus);
    *rbValue = realnbins;
	dblNbins = (epicsFloat64)*rbValue;

	for(i = 0; i < this->nChannels; i++)
	{
		xiastatus = xiaSetAcquisitionValues(i, "number_mca_channels", (void*)&dblNbins);
		status = this->xia_checkError(pasynUser, xiastatus, "number_mca_channels");
		if (status == asynError)
			{
			asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR, "%s::%s [%s] can not set nbins=%u (%.3f) ch=%u\n",
						driverName, functionName, this->portName, *rbValue, dblNbins, i);
			return status;
			}
	}

	/* Finally apply the operation to the board in order for it to take effect */
	for (i=0; i < this->nChannels; i+=XMAP_NCHANS_MODULE)
	{
		xiastatus = xiaBoardOperation(i, "apply", (void*)&dummy);
		status = this->xia_checkError(pasynUser, xiastatus, "apply");
		if (status == asynError) return status;
		if (mode == ADX_mapping)
		{
			setIntegerParam(i, ADImageSize, (int)((*rbValue)*sizeof(unsigned long)));
			callParamCallbacks(i, i);
		}
	}

	/* If in mapping mode we need to modify the Y size as well in order to fit in the 2MB buffer!
	 * We also need to read out the new lenght of the mapping buffer because that can possibly change... */
	getIntegerParam(ADX_xmap_mode, &mode);
	if (mode == ADX_mapping)
	{
		for (i = 0; i < this->nChannels; i+=XMAP_NCHANS_MODULE)
		{
			xiastatus = xiaSetAcquisitionValues(i, "num_map_pixels_per_buffer", (void*)&dNumPixPerBuf);
			status = this->xia_checkError(pasynUser, xiastatus, "num_map_pixels_per_buffer");
			xiastatus = xiaBoardOperation(i, "apply", (void*)&dummy);
			status = this->xia_checkError(pasynUser, xiastatus, "apply");

			xiastatus = xiaGetAcquisitionValues(i, "num_map_pixels_per_buffer", (void*)&dTmp);
			status = this->xia_checkError(pasynUser, xiastatus, "GET num_map_pixels_per_buffer");
			maxSizeY = (int)dTmp;
			asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s [%d] Got num_map_pixels_per_buffer = %d\n", driverName, functionName, i, maxSizeY);

			bufLen = 0;
			xiastatus = xiaGetRunData(i, "buffer_len", (void*)&bufLen);
			status = this->xia_checkError(pasynUser, xiastatus, "GET buffer_len");
			asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s [%d] Got buffer_len = %u\n", driverName, functionName, i, bufLen);
			setIntegerParam(i, ADImageSize, (int)(bufLen*sizeof(unsigned long)));
			callParamCallbacks(i, i);
		}
		setIntegerParam(ADSizeY, maxSizeY);
		setIntegerParam(ADMaxSizeY, maxSizeY);
		callParamCallbacks();
	}

	return status;
}

asynStatus ADXmap::changeMode(asynUser *pasynUser, epicsInt32 mode)
{
	asynStatus status = asynSuccess;
	int oldmode, xiastatus;
	unsigned int i;
	epicsUInt32 firstCh, bufLen;
	double dummy, dMode=0, dTmp, dNumPixPerBuf = -1.0;
	const char* functionName = "changeMode";

	if (mode < 0 || mode > 1) return asynError;

	/* TODO: Check if we are running right now.... */

	getIntegerParam(ADX_xmap_mode, &oldmode);
	if (oldmode == mode) return asynSuccess;

	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s Changing mode to %d (%.1f)\n",driverName, functionName, mode, dMode);

	dMode = (double)mode;
	setIntegerParam(ADStatus, 7);
	callParamCallbacks();
	xiastatus = xiaSetAcquisitionValues(XMAP_ALL, "mapping_mode", (void*)&dMode);
	status = this->xia_checkError(pasynUser, xiastatus, "mapping_mode");
	setIntegerParam(ADStatus, 0);
	callParamCallbacks();
	if (status == asynError) return status;

	switch(mode)
	{
	case ADX_normal:
		setIntegerParam(ADSizeY, 1);
		setIntegerParam(ADMaxSizeY, 1);
		getIntegerParam(ADSizeX, (int*)&bufLen);
		setIntegerParam(ADDataType, NDUInt32);
		dummy = 0.0;
		for (i=0; i < this->nChannels; i++)
		{
			/* Clear the normal mapping mode status settings */
			setIntegerParam(i, ADX_stat_events, 0);
			setDoubleParam(i, ADX_stat_inputcountrate, dummy);
			setDoubleParam(i, ADX_stat_outputcountrate, dummy);

			/* Set the new image size down to one array of sizeX x epicsUInt32 words */
			setIntegerParam(i, ADImageSize, (int)(bufLen*sizeof(epicsUInt32)));
			callParamCallbacks(i,i);
		}
		break;
	case ADX_mapping:
		int maxSizeY = 0;
		for (firstCh = 0; firstCh < this->nChannels; firstCh+=XMAP_NCHANS_MODULE)
		{
			xiastatus = xiaSetAcquisitionValues(firstCh, "num_map_pixels_per_buffer", (void*)&dNumPixPerBuf);
			status = this->xia_checkError(pasynUser, xiastatus, "num_map_pixels_per_buffer");
			xiastatus = xiaBoardOperation(firstCh, "apply", (void*)&dummy);
			status = this->xia_checkError(pasynUser, xiastatus, "apply");

		}

		for (firstCh = 0; firstCh < this->nChannels; firstCh+=XMAP_NCHANS_MODULE)
		{
			dTmp = 0;
			xiastatus = xiaGetAcquisitionValues(firstCh, "num_map_pixels_per_buffer", (void*)&dTmp);
			status = this->xia_checkError(pasynUser, xiastatus, "GET num_map_pixels_per_buffer");
			asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s [%d] Got num_map_pixels_per_buffer = %.1f\n", driverName, functionName, firstCh, dTmp);
			maxSizeY = (int)dTmp;

			bufLen = 0;
			xiastatus = xiaGetRunData(firstCh, "buffer_len", (void*)&bufLen);
			status = this->xia_checkError(pasynUser, xiastatus, "GET buffer_len");
			asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s [%d] Got buffer_len = %u\n", driverName, functionName, firstCh, bufLen);
			setIntegerParam(firstCh, ADImageSize, (int)(bufLen*sizeof(epicsUInt16)));
			callParamCallbacks(firstCh, firstCh);

			/* Clear the normal mapping mode status settings */
			dummy = 0.0;
			for (i=0; i < XMAP_NCHANS_MODULE; i++)
			{
				setIntegerParam(firstCh+i, ADX_stat_triggers, 0);
				setDoubleParam(firstCh+i, ADX_stat_realtime, dummy);
				setDoubleParam(firstCh+i, ADX_stat_triggerlivetime, dummy);
				setDoubleParam(firstCh+i, ADX_stat_energylivetime, dummy);
				callParamCallbacks(firstCh+i, firstCh+i);
			}
		}

		setIntegerParam(ADSizeY, maxSizeY);
		setIntegerParam(ADMaxSizeY, maxSizeY);
		setIntegerParam(ADDataType, NDUInt16);
		break;
	}

	return status;
}

asynStatus ADXmap::calcMaxY(asynUser *pasynUser)
{
	asynStatus status = asynSuccess;
	return status;
}


/** Create an asyn user for the driver.
 * Maps the integer/enum asyn commands on to a string representation that
 * can be used to indicate a certain command in in the INP/OUT field of a record.
 * \param pasynUser
 * \param drvInfo
 * \param pptypeName
 * \param psize
 * \return asynStatus Either asynError or asynSuccess
 */
asynStatus ADXmap::drvUserCreate( asynUser *pasynUser, const char *drvInfo, const char **pptypeName, size_t *psize)
{
	asynStatus status;
	int param;
	const char *functionName = "drvUserCreate";

	/* See if this is one of the drivers local parameters */
	status = findParam(ADXParamString, ADX_N_PARAMS, drvInfo, &param);

	if (status == asynSuccess)
	{
		pasynUser->reason = param;
		if (pptypeName) { *pptypeName = epicsStrDup(drvInfo); }
		if (psize) { *psize = sizeof(param); }
		asynPrint(	pasynUser, ASYN_TRACE_FLOW, "%s:%s: drvInfo=%s, param=%d\n",
					driverName, functionName, drvInfo, param);
		return asynSuccess;
	}

	/* If not a local driver parameter, then see if it is a base class parameter */
	status = ADDriver::drvUserCreate(pasynUser, drvInfo, pptypeName, psize);
	return status;
}

/** Thread used to poll the hardware for status and data.
 *
 */
void ADXmap::acquisitionTask()
{
	asynStatus status = asynSuccess;
	int paramStatus, xiastatus;
	int numImages, imageCounter, numImagesCounter, acquire, mode;
	int externalStopCmd = 0, imageMode, resume = 0;
	epicsFloat64 acqperiod, sleeptime, dtmp;
	epicsTimeStamp now, start;
	unsigned int firstCh;
	const char* functionName = "acquisitionTask";

	setDoubleParam(ADAcquirePeriod, 0.1);
	asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW, "acquisition task started!\n");
	epicsEventTryWait(this->stopEventId); /* clear the stop event if it wasn't already */
	epicsMutexLock(this->mutexId);

	while (1) /* ... round and round and round we go ... */
	{
		/* Is acquisition active? */
		getIntegerParam(ADAcquire, &acquire);

		/* If we are not acquiring then wait for a semaphore that is given when acquisition is started */
		if (!acquire)
		{
			asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW, "acquireTask: ADAcquire=%d\n", acquire);
			setIntegerParam(ADStatus, ADStatusIdle);
			callParamCallbacks();

			if (externalStopCmd)
			{
				/* Signal someone that the thread is really stopping to wait for start event */
				asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW,
						"%s::%s [%s]: Signalling stop event\n", driverName, functionName, this->portName);
				epicsEventSignal(this->stopEventId);
			} else externalStopCmd = 1;

			/* XIA Stop command here.... */
			xiastatus = xiaStopRun(XMAP_ALL);		/* Stop the run */
			status = this->xia_checkError( this->pasynUserSelf, xiastatus, "xiaStopRun()" );
			if (status == asynError)
			{
				asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR, "%s::%s [%s] Stopping acquisition failed...\n",
							driverName, functionName, this->portName);
			}

			/* Release the lock while we wait for an event that says acquire has started, then lock again */
			epicsMutexUnlock(this->mutexId);

			/* Wait for a signal that tells this thread that the transmission
			 * has started and we can start asking for data...	 */
			asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW,
					"%s::%s [%s]: waiting for acquire to start\n", driverName, functionName, this->portName);
			epicsEventWait(this->startEventId);
			epicsMutexLock(this->mutexId);

			/* make sure we use buffer A to start with */
			for (firstCh=0; firstCh < this->nChannels; firstCh += XMAP_NCHANS_MODULE) this->currentBuf[firstCh] = 0;
			/* Start the run!  */
			paramStatus |= getIntegerParam(ADX_start_mode, &resume);
			xiastatus = xiaStartRun(XMAP_ALL, resume);
			status = this->xia_checkError( this->pasynUserSelf, xiastatus, "xiaStartRun()" );
			if (status == asynError) continue; /* TODO: some better error handling if starting fails... */
			paramStatus  = getIntegerParam(ADX_xmap_mode, &mode);
			asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW,
					"%s::%s [%s]: started! (mode=%d)\n", driverName, functionName, this->portName, mode);
			setIntegerParam(ADNumImagesCounter, 0);
			setIntegerParam(ADAcquire, TRUE);
			setDoubleParam(ADX_mbytes_received, 0.0);
		}
		epicsTimeGetCurrent(&start);

		/* First get and update the status of the XMAP */
		this->getStatus(this->pasynUserSelf);

		/* Check what mode we are in (mca or mapping) */
		switch (mode)
		{
		case 0:
			/* Get the data */
			status = this->pollMcaMode();
			break;
		case 1:
			status = this->pollMappingMode();
			break;
		default:
			break;
		}

		/* See if acquisition is done if we are in single or multiple mode */
		paramStatus |= getIntegerParam(ADNumImages, &numImages);
		paramStatus |= getIntegerParam(ADImageMode, &imageMode);
		paramStatus |= getIntegerParam(ADImageCounter, &imageCounter);
		paramStatus |= getIntegerParam(ADNumImagesCounter, &numImagesCounter);
		if ((imageMode == ADImageSingle) ||
			((imageMode == ADImageMultiple) && (numImagesCounter >= numImages)))
		{
			/* command the camera to stop acquiring.. */
			setIntegerParam(ADAcquire, 0);
			externalStopCmd = 0;
		}
		paramStatus |= getDoubleParam(ADAcquirePeriod, &acqperiod);
		epicsTimeGetCurrent(&now);
		dtmp = epicsTimeDiffInSeconds(&now, &start);
		sleeptime = acqperiod - dtmp;
		if (sleeptime > 0.0)
		{
			epicsMutexUnlock(this->mutexId);
			epicsThreadSleep(sleeptime);
			epicsMutexLock(this->mutexId);
		}
	}
}

asynStatus ADXmap::pollMcaMode()
{
	asynStatus status = asynSuccess;
	unsigned int i;
	/*unsigned int tmp, tmpp, index;*/
	/*int c;*/
	int ch, xiastatus, paramStatus;
	int nBins=0, arrayCallbacks;
	int dims[2], imageCounter, numImagesCounter;
	NDDataType_t dataType;
	epicsTimeStamp now;
	//const char* functionName = "pollMcaMode";

	paramStatus  = getIntegerParam(ADSizeX, &nBins);
	paramStatus |= getIntegerParam(ADArrayCallbacks, &arrayCallbacks);
	paramStatus |= getIntegerParam(ADDataType, (int *)&dataType);
	paramStatus |= getIntegerParam(ADSizeX, &dims[0]);
	paramStatus |= getIntegerParam(ADSizeY, &dims[1]);
	paramStatus |= getIntegerParam(ADImageCounter, &imageCounter);
	paramStatus |= getIntegerParam(ADNumImagesCounter, &numImagesCounter);

	epicsTimeGetCurrent(&now);
	imageCounter++;
	numImagesCounter++;
	setIntegerParam(ADImageCounter, imageCounter);
	setIntegerParam(ADNumImagesCounter, numImagesCounter);
	callParamCallbacks();

	/* Run through all boards and collect the spectrums */
	for (i=0; i < this->nChannels; i+=XMAP_NCHANS_MODULE)
	{
		/* Collect all channels on the current board into a temporary buffer */
		epicsMutexUnlock(this->mutexId);
		xiastatus = xiaGetRunData(i, "module_mca", (void*)this->pMcaRaw);
		status = this->xia_checkError(this->pasynUserSelf, xiastatus, "xiaGetRunData()");
		epicsMutexLock(this->mutexId);
		if (status != asynSuccess) return status;

		/* Run through the 4 channels on each board... */
		for (ch=0; ch < XMAP_NCHANS_MODULE; ch++)
		{
			/* Only bother to fill up the callback buffer if we are requesting callbacks... */
			if (arrayCallbacks)
			{

				/* Allocate one buffer for callback
				 * -and just telling it where to find the data in the pMcaRaw buffer... */
				this->pRaw = this->pNDArrayPool->alloc(2, dims, dataType, 0, (void*)(this->pMcaRaw+(ch*nBins)) );
				this->pRaw->timeStamp = now.secPastEpoch + now.nsec / 1.e9;
				this->pRaw->uniqueId = imageCounter;
				epicsMutexUnlock(this->mutexId);
				doCallbacksGenericPointer(this->pRaw, NDArrayData, i+ch);
				epicsMutexLock(this->mutexId);
				this->pRaw->release();
			}

		}
	}
	return status;
}

asynStatus ADXmap::pollMappingMode()
{
	asynStatus status = asynSuccess;
	int xiastatus;
	unsigned int ch;
	int buf = 0, nWords, i;
	unsigned short isfull;
	epicsUInt32 *data = this->pMapRaw;
	int paramStatus=asynSuccess, arrayCallbacks, dims[2], imageCounter, numImagesCounter, bufSize;
	NDDataType_t dataType;
	epicsTimeStamp now, after;
	double readoutTime, readoutBurstSpeed, mbytesRead, MBbufSize;
	const char* functionName = "pollMappingMode";

	paramStatus |= getIntegerParam(ADDataType, (int *)&dataType);
	paramStatus |= getIntegerParam(ADImageCounter, &imageCounter);
	paramStatus |= getIntegerParam(ADNumImagesCounter, &numImagesCounter);
	paramStatus |= getIntegerParam(ADSizeX, &dims[0]);
	paramStatus |= getIntegerParam(ADSizeY, &dims[1]);
	paramStatus |= getIntegerParam(ADImageSize, &bufSize);
	paramStatus |= getDoubleParam(ADX_mbytes_received, &mbytesRead);
	MBbufSize = (double)(bufSize) / (double)MEGABYTE;

	/* Step through all the first channels on all the boards in the system. */
	for (ch=0; ch < this->nChannels; ch+=XMAP_NCHANS_MODULE)
	{
		buf = this->currentBuf[ch];

		xiastatus = xiaGetRunData(ch, ADXBuffers[buf].bufFull, (void*)&isfull);
		status = this->xia_checkError(this->pasynUserSelf, xiastatus, ADXBuffers[buf].bufFull);
		/*asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s %s isfull=%d\n",
					driverName, functionName, ADXBuffers[buf].bufFull, isfull);*/
		if (isfull)
		{
			epicsTimeGetCurrent(&now);
			imageCounter++; numImagesCounter++;

			/* the buffer is full so read it out */
			xiastatus = xiaGetRunData(ch, ADXBuffers[buf].buffer, (void*)data);
			epicsTimeGetCurrent(&after);
			status = this->xia_checkError(this->pasynUserSelf, xiastatus, ADXBuffers[buf].buffer);
			paramStatus |= getIntegerParam(ADArrayCallbacks, &arrayCallbacks);

			readoutTime = epicsTimeDiffInSeconds(&after, &now);
			readoutBurstSpeed = MBbufSize / readoutTime;
			mbytesRead += MBbufSize;
			asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW, "%s::%s Got data! size=%.3fMB (%d) dt=%.3fs speed=%.3fMB/s\n",
					driverName, functionName, MBbufSize, bufSize, readoutTime, readoutBurstSpeed);

			if (arrayCallbacks)
			{
				//this->pRaw = this->pNDArrayPool->alloc(2, dims, dataType, 0, (void*)data );
				this->pRaw = this->pNDArrayPool->alloc(2, dims, dataType, bufSize, NULL );

				/* First get rid of the empty parts of the 32 bit words. The Handel library
				 * provides a 4MB 32 bit/word wide buffer -even though the data only takes up
				 * 2MB 16bit word size. So we get rid of the empty parts here. */
				nWords = bufSize / sizeof(epicsUInt16);
				for(i = 0; i < nWords; i++)
				{
					*(((epicsUInt16*)this->pRaw->pData)+i) = (epicsUInt16)*(data + i);
				}

				this->pRaw->timeStamp = now.secPastEpoch + now.nsec / 1.e9;
				this->pRaw->uniqueId = imageCounter;
				this->pRaw->dataSize = bufSize;
				epicsMutexUnlock(this->mutexId);
				doCallbacksGenericPointer(this->pRaw, NDArrayData, ch);
				epicsMutexLock(this->mutexId);
				this->pRaw->release();

				setDoubleParam(ADX_mbytes_received, mbytesRead);
				setDoubleParam(ADX_read_speed, readoutBurstSpeed);
				callParamCallbacks();
			}
			/*asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW, "%s::%s Done reading! Ch=%d bufchar=%s\n",
					driverName, functionName, ch, ADXBuffers[buf].bufChar);*/
			/* Notify XMAP that we read out the buffer */
			xiastatus = xiaBoardOperation(ch, "buffer_done", (void*)ADXBuffers[buf].bufChar);
			status = this->xia_checkError(this->pasynUserSelf, xiastatus, "buffer_done");
			if (buf == 0) this->currentBuf[ch] = 1;
			else this->currentBuf[ch] = 0;
		}
	}
	return status;
}

asynStatus ADXmap::getStatus(asynUser* pasynUser)
{
	asynStatus status = asynSuccess;
	epicsInt32 xiastatus;
	unsigned int i, acqState, firstCh, ch, currentPixel;
	int iTmpVal, mode, acquiring;
	unsigned short bufferOverrun;
	//const char* functionName = "getStatus";
	double *tmpStatModule;
	//double dTmp;

	xiastatus = xiaGetRunData(0, "run_active", (void*)&acqState);
	status = this->xia_checkError(pasynUser, xiastatus, "run_active");
	if (status == asynError) return status;
	setIntegerParam(ADX_xmap_runstate, acqState);
	callParamCallbacks();

	getIntegerParam(ADX_xmap_mode, &mode);
	getIntegerParam(ADAcquire, &acquiring);
	for (firstCh=0; firstCh<this->nChannels ;firstCh+=XMAP_NCHANS_MODULE)
	{
		switch(mode)
		{
			case ADX_normal:

				/* Get the runtime stats like livetime and so on... */
				xiastatus = xiaGetRunData(firstCh, "module_statistics", (void*)this->tmpStats);
				status = this->xia_checkError(pasynUser, xiastatus, "module_statistics");
				if (status == asynError) return status;

				for (i=0; i < XMAP_NCHANS_MODULE; i++)
				{
					ch = firstCh+i;
					tmpStatModule = this->tmpStats + (i*7);
					setDoubleParam(ch, ADX_stat_realtime, tmpStatModule[0]);
					setDoubleParam(ch, ADX_stat_triggerlivetime, tmpStatModule[1]);
					setDoubleParam(ch, ADX_stat_energylivetime, tmpStatModule[2]);
					iTmpVal = (int)tmpStatModule[3];
					setIntegerParam(ch, ADX_stat_triggers, iTmpVal);
					iTmpVal = (int)tmpStatModule[4];
					setIntegerParam(ch, ADX_stat_events, iTmpVal);
					setDoubleParam(ch, ADX_stat_inputcountrate, tmpStatModule[5]/1000.0);
					setDoubleParam(ch, ADX_stat_outputcountrate, tmpStatModule[6]/1000.0);
					callParamCallbacks(ch, ch);
				}
				break;

			case ADX_mapping:
				xiastatus = xiaGetRunData(firstCh, "current_pixel", (void*)&currentPixel);
				status = this->xia_checkError(pasynUser, xiastatus, "current_pixel");
				setIntegerParam(firstCh, ADX_current_pixel ,currentPixel);
				xiastatus = xiaGetRunData(firstCh, "buffer_overrun", (void*)&bufferOverrun);
				status = this->xia_checkError(pasynUser, xiastatus, "buffer_overrun");
				setIntegerParam(firstCh, ADX_buffer_overrun ,(int)bufferOverrun);

				/*for (i=0; i < XMAP_NCHANS_MODULE; i++)
				{
					ch = firstCh+i;
					xiastatus = xiaGetRunData(ch, "real_time", (void*)&dTmp);
					status = this->xia_checkError(pasynUser, xiastatus, "real_time");
					setDoubleParam(ch, ADX_stat_realtime, dTmp);

					xiastatus = xiaGetRunData(ch, "trigger_livetime", (void*)&dTmp);
					status = this->xia_checkError(pasynUser, xiastatus, "trigger_livetime");
					setDoubleParam(ch, ADX_stat_triggerlivetime, dTmp);

					xiastatus = xiaGetRunData(ch, "livetime", (void*)&dTmp);
					status = this->xia_checkError(pasynUser, xiastatus, "livetime");
					setDoubleParam(ch, ADX_stat_energylivetime, dTmp);

					callParamCallbacks(ch, ch);
				}*/
				break;
		}
	}

	return status;
}


void ADXmap::report(FILE *fp, int details)
{


}


asynStatus ADXmap::err( asynUser* asynUser, int errOriginLine)
{
	asynStatus status = asynSuccess;

	return status;
}

asynStatus ADXmap::xia_checkError( asynUser* pasynUser, epicsInt32 xiastatus, char *xiacmd )
{
	if (xiastatus == XIA_SUCCESS) return asynSuccess;

	asynPrint( pasynUser, ASYN_TRACE_ERROR, "### ADXmap: XIA ERROR: %d (%s)\n", xiastatus, xiacmd );
	return asynError;
}

void ADXmap::shutdown()
{
	int status;
	asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW, "%s: shutting down...\n", driverName);
	status = xiaExit();
	if (status == XIA_SUCCESS)
	{
		printf("### XIA XMAP closed down successfully. You can now exit the iocshell ###\n");
		return;
	}
	printf("xiaExit() error: %d\n", status);
	return;
}
