/* Standard includes... */
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* EPICS includes */
#include <epicsString.h>
#include <epicsTime.h>
#include <epicsThread.h>
#include <epicsEvent.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <epicsExit.h>
#include <envDefs.h>


/* Area Detector includes */
#include <NDArray.h>
#include <ADDriver.h>
#include <NDPluginDriver.h>
#include "adXmap.h"

typedef enum ADXmapFileParam_t {
	XMF_file = NDPluginDriverLastParam, /** <  */
	XMF_bytes_written,
	XMF_burst_write_speed,
	XMF_avg_write_speed,

	XMFLastParam
} ADXmapFileParam_t;

static asynParamString_t ADXmapFileParamString[] = {
		{XMF_file,                "XMF_FILE"},
		{XMF_bytes_written,       "XMF_BYTES_WRITTEN"},
		{XMF_burst_write_speed,   "XMF_BURST_WRITE_SPEED"},
		{XMF_avg_write_speed,     "XMF_AVG_WRITE_SPEED"},
};

/** Number of asyn parameters (asyn commands) this driver supports. */
#define ADXMAPFILE_N_PARAMS (sizeof( ADXmapFileParamString)/ sizeof(ADXmapFileParamString[0]))
#define MAX_FILENAME_SIZE 256


static ENV_PARAM XIAHOME;

static const char * driverName = "adXmapMapFile";

class ADXmapMapFile : public NDPluginDriver
{
public:
	ADXmapMapFile(const char *portName, int queueSize, const char *NDArrayPort, int xmapChannel, int maxBuf, int maxMem);
	/* These methods override those in the base class */
	virtual void processCallbacks(NDArray *pArray);
	virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
	virtual asynStatus drvUserCreate(asynUser *pasynUser, const char *drvInfo, const char **pptypeName, size_t *psize);
private:
	asynStatus openFile(asynUser *pasynUser, int open);
	FILE *fptr;
	char * filename;
	epicsUInt16 *dataBuf;
	epicsTimeStamp tStartAcq;
};


extern "C" int ADXmap_MapFileConfig(const char *portName, int queueSize, const char *NDArrayPort, int xmapChannel, int maxBuf, int maxMem)
{
    new ADXmapMapFile(portName, queueSize, NDArrayPort, xmapChannel, maxBuf, maxMem);
    return(asynSuccess);
}

ADXmapMapFile::ADXmapMapFile(const char* portName, int queueSize, const char* NDArrayPort, int xmapChannel, int maxBuf, int maxMem)
: NDPluginDriver(portName, queueSize, 0, NDArrayPort, xmapChannel, 1, XMFLastParam, maxBuf, maxMem,
		asynInt32ArrayMask | asynGenericPointerMask | asynOctetMask | asynInt32Mask,
		asynInt32ArrayMask | asynGenericPointerMask | asynOctetMask | asynInt32Mask,
		ASYN_MULTIDEVICE | ASYN_CANBLOCK, 0, 0, 0)
{
	char xia_home_path[MAX_FILENAME_LEN];
	char *fileprefix, *filetemplate;

	this->fptr = NULL;
	this->filename = (char*)calloc(MAX_FILENAME_SIZE, sizeof(char));
	this->dataBuf = (epicsUInt16*) calloc(XMAP_MAX_MAPBUF_SIZE, sizeof(epicsUInt16));


	/* Try to connect to the array port */
	connectToArrayPort();

	/* Get the XIAHOME environment variable to define where to write the
	 * error log files to.		 */
	XIAHOME.name = epicsStrDup("XIAHOME");
	XIAHOME.pdflt = (char*)calloc(MAX_FILENAME_LEN, sizeof(char));
	envGetConfigParam(&XIAHOME, MAX_FILENAME_LEN, xia_home_path);

	setStringParam(NDFilePath, xia_home_path);
	fileprefix = epicsStrDup("prefix");
	setStringParam(NDFileName, fileprefix);
	filetemplate = epicsStrDup("%s%s_%d");
	setStringParam(NDFileTemplate, filetemplate);
	callParamCallbacks();
}

void ADXmapMapFile::processCallbacks(NDArray *pArray)
{
	int nWords, nWordsWritten, nBytesWritten;
	double MBytesWritten, tmpMBytesWritten;
	int arrayCounter;
	//epicsUInt32 *tmpRawData;
	epicsFloat64 avgWriteSpeed;
	const char* functionName = "processCallbacks";
	epicsTimeStamp start, end;
	epicsFloat64 writeTime, writeSpeed;

	getDoubleParam(XMF_avg_write_speed, &avgWriteSpeed);
	if (avgWriteSpeed == 0.0) epicsTimeGetCurrent(&this->tStartAcq);

	/* Most plugins want to increment the arrayCounter each time they are called, which NDPluginDriver
	 * does.  However, for this plugin we only want to increment it when we actually got a callback we were
	 * supposed to save.  So we save the array counter before calling base method, increment it here */
	getIntegerParam(NDPluginDriverArrayCounter, &arrayCounter);
	arrayCounter++;

	/* Call the base class method */
	NDPluginDriver::processCallbacks(pArray);

	/* Reserve an array to work with */
	pArray->reserve();

	nWords = pArray->dataSize/sizeof(epicsUInt16);
	asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW, "%s::%s ID=%u, nWords=%u, bytesize=%u, datatype=%d\n",
			driverName, functionName, pArray->uniqueId, nWords, pArray->dataSize, pArray->dataType);

	getDoubleParam(XMF_bytes_written, &MBytesWritten);

	/* Dump the data to the file here */
	epicsTimeGetCurrent(&start);
	if (this->fptr != NULL)
	{
		nWordsWritten = fwrite((void*)pArray->pData, sizeof(epicsUInt16), nWords, this->fptr);
		fflush(this->fptr);
		epicsTimeGetCurrent(&end);

		nBytesWritten = nWordsWritten * sizeof(epicsUInt16);
		tmpMBytesWritten = (double)nBytesWritten / (double)MEGABYTE;
		writeTime = epicsTimeDiffInSeconds(&end, &start);
		writeSpeed = tmpMBytesWritten / writeTime;

		asynPrint(this->pasynUserSelf, ASYN_TRACE_FLOW, "%s::%s Filedump done! Written stats: bytes=%d words=%d MB=%.3f t=%.3fs speed=%.3fMB/s\n",
				driverName, functionName, nBytesWritten, nWordsWritten, tmpMBytesWritten, writeTime, writeSpeed);

		MBytesWritten += tmpMBytesWritten;
		setDoubleParam(XMF_bytes_written, MBytesWritten);

		avgWriteSpeed = MBytesWritten/epicsTimeDiffInSeconds(&end, &this->tStartAcq);
		setDoubleParam(XMF_avg_write_speed, avgWriteSpeed);
		setDoubleParam(XMF_burst_write_speed, writeSpeed);
		callParamCallbacks();

	} else
	{
		asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR, "%s::%s ERROR: no file buffer available!\n", driverName, functionName);
	}

	/* now that we copied the data into the 16 bit buffer, we can release the original one */
	pArray->release();

	/* Update the parameters.  */
	setIntegerParam(NDPluginDriverArrayCounter, arrayCounter);
	callParamCallbacks();
}

asynStatus ADXmapMapFile::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	int rbValue;
	//const char *functionName = "writeInt32";

	rbValue = value;

	switch(pasynUser->reason)
	{
	case XMF_file:
		status = this->openFile(pasynUser, value);
		break;
	default:
		/* This was not a parameter that this driver understands, try the base class */
		status = NDPluginDriver::writeInt32(pasynUser, value);
		break;
	}

	if (status == asynSuccess)
	{
		setIntegerParam(pasynUser->reason, rbValue);
		callParamCallbacks();
	}
	return status;
}

asynStatus ADXmapMapFile::openFile(asynUser *pasynUser, int open)
{
	asynStatus status = asynSuccess;
	char filePath[MAX_FILENAME_LEN];
	char fileName[MAX_FILENAME_LEN];
	char fileTemplate[MAX_FILENAME_LEN];
	int fileNumber, autoIncrement;
	const char * functionName = "openFile";


	if (open == FALSE)
	{
		if (this->fptr != NULL) fclose(this->fptr);
		this->fptr = NULL;
		return status;
	}

	setDoubleParam(XMF_avg_write_speed, 0.0);
	setDoubleParam(XMF_bytes_written, 0.0);
	setDoubleParam(XMF_burst_write_speed, 0.0);

	getStringParam(NDFilePath, sizeof(filePath), filePath);
	getStringParam(NDFileName, sizeof(fileName), fileName);
	getStringParam(NDFileTemplate, sizeof(fileTemplate), fileTemplate);
	getIntegerParam(NDFileNumber, &fileNumber);
	getIntegerParam(NDAutoIncrement, &autoIncrement);

	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s Got filepath = \"%s\"\n",
			driverName, functionName, filePath);
	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s Got filename = \"%s\"\n",
			driverName, functionName, fileName);
	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s Got filetemplate = \"%s\"\n",
			driverName, functionName, fileTemplate);
	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s AutoInc=%d File # = %d\n",
			driverName, functionName, autoIncrement, fileNumber);


	this->createFileName(MAX_FILENAME_LEN, this->filename);
	asynPrint(pasynUser, ASYN_TRACE_FLOW, "%s::%s Created filename: \"%s\"\n",
			driverName, functionName, this->filename);
	setStringParam(NDFullFileName, this->filename);
	//callParamCallbacks();

	this->fptr = fopen(this->filename, "wb");
	if (this->fptr == NULL)
	{
		asynPrint(pasynUser, ASYN_TRACE_ERROR, "%s::%s ERROR: Could not open file \"%s\"\n", driverName, functionName, this->filename);
		status = asynError;
	}

	return status;
}

asynStatus ADXmapMapFile::drvUserCreate(asynUser *pasynUser, const char* drvInfo, const char **pptypeName, size_t *psize)
{
	asynStatus status;
	int param;

	/* Look in the driver table */
	status = findParam(ADXmapFileParamString, ADXMAPFILE_N_PARAMS, drvInfo, &param);
	if (status == asynSuccess)
	{
		pasynUser->reason = param;
		if (pptypeName) *pptypeName = epicsStrDup(drvInfo);
		if (psize) *psize = sizeof(param);
		asynPrint(pasynUser, ASYN_TRACE_FLOW,
				"%s::drvUserCreate, drvInfo=%s, param=%d\n",
				driverName, drvInfo, param);
		return(asynSuccess);
	}

	/* If we did not find it in that table try the plugin base */
	status = NDPluginDriver::drvUserCreate(pasynUser, drvInfo, pptypeName, psize);
	return(status);
}

