/*******************/
/* System includes */
/*******************/
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>

/******************/
/* EPICS includes */
/******************/

/* asyn includes first */
/*#include <asynDriver.h>*/
#include <asynInt32.h>
#include <asynFloat64.h>
#include <asynInt32Array.h>
#include <asynUInt32Digital.h>
#include <asynDrvUser.h>

/* rest of epics includes */
#include <cantProceed.h>	/* provides: callocMustSucceed() */
/*#include <devLib.h>*/
#include <epicsMutex.h>
#include <epicsString.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <epicsExit.h>
#include <envDefs.h>

#include "drvXmap.h"

/*********************************************************/
/*                                                      */
/* Asyn interface Structures and Function prototypes   */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
static void xmap_Report                     (void *drvPvt, FILE *fp, int details);
static asynStatus xmap_Connect              (void *drvPvt, asynUser *pasynUser);
static asynStatus xmap_Disconnect           (void *drvPvt, asynUser *pasynUser);

/* asynInt32 methods */
static asynStatus xmap_int32Write           (void *drvPvt, asynUser *pasynUser, epicsInt32 value);
static asynStatus xmap_int32Read            (void *drvPvt, asynUser *pasynUser, epicsInt32 *value);
static asynStatus xmap_getBounds            (void *drvPvt, asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high);

/* asynUInt32Digital methods */
static asynStatus xmap_uint32DigitalWrite   (void *drvPvt, asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask);
static asynStatus xmap_uint32DigitalRead    (void *drvPvt, asynUser *pasynUser, epicsUInt32 *value, epicsUInt32 mask);

/* asynFloat64 methods */
static asynStatus xmap_float64Write         (void *drvPvt, asynUser *pasynUser, epicsFloat64 value);
static asynStatus xmap_float64Read          (void *drvPvt, asynUser *pasynUser, epicsFloat64 *value);

/* asynInt32Array methods */
static asynStatus xmap_int32ArrayRead       (void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t maxChans, size_t *nactual);
static asynStatus xmap_int32ArrayWrite      (void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t maxChans);

/* asynDrvUser methods */
static asynStatus xmap_drvUserCreate         (void *drvPvt, asynUser *pasynUser, const char *drvInfo, const char **pptypeName, size_t *psize);
static asynStatus xmap_drvUserGetType        (void *drvPvt, asynUser *pasynUser, const char **pptypeName, size_t *psize);
static asynStatus xmap_drvUserDestroy        (void *drvPvt, asynUser *pasynUser);

/* local functions */
xmap_command_t xmap_get_cmd( asynUser* pasynUser );
asynStatus xmap_get_trace(	asynUser* pasynUser, xmap_driver_t *pPvt,
		epicsInt32* data, epicsInt32 maxlen, epicsInt32* actuallen);

static struct asynCommon xmap_Common =        { xmap_Report, xmap_Connect, xmap_Disconnect};
static asynInt32 xmap_Int32 =                 { xmap_int32Write, xmap_int32Read, xmap_getBounds, NULL, NULL};
static asynUInt32Digital xmap_UInt32Digital = { xmap_uint32DigitalWrite, xmap_uint32DigitalRead, NULL, NULL, NULL, NULL, NULL};
static asynFloat64 xmap_Float64 =             { xmap_float64Write, xmap_float64Read, NULL, NULL};
static asynInt32Array xmap_Int32Array =       { xmap_int32ArrayWrite, xmap_int32ArrayRead, NULL, NULL};
static asynDrvUser xmap_DrvUser =             { xmap_drvUserCreate, xmap_drvUserGetType, xmap_drvUserDestroy};

static xmap_command_t xmap_commands[XMAP_N_CMDS] = {
		{xmap_mca_collect,      "xmap_mca_collect",     NULL, XMAP_UINT32},
		{xmap_mca_stat,         "xmap_mca_stat",        NULL, XMAP_UINT32},
		{xmap_mca_triggers,     "xmap_mca_triggers",    NULL, XMAP_UINT32},
		{xmap_mca_preset_type,  "xmap_mca_preset_type", NULL, XMAP_UINT32},
		{xmap_mca_preset_value, "xmap_mca_preset_value",NULL, XMAP_FLOAT64},
		{xmap_mca_acqstat,      "xmap_mca_acqstat",     NULL, XMAP_INT32},
		{xmap_start_all,        "xmap_start_all",       NULL, XMAP_UINT32},
		{xmap_some_command,     "xmap_some_command",    NULL, XMAP_INT32},
		{xmap_another_command,  "xmap_another_command", NULL, XMAP_INT32},
		{xmap_run,              "xmap_run",             NULL, XMAP_UINT32},
		{xmap_trace,	            "xmap_trace",           NULL, XMAP_INT32ARRAY|XMAP_FLOAT64|XMAP_UINT32},
		{xmap_enable_ch,        "xmap_enable_ch",       NULL, XMAP_UINT32},
		{xmap_enable_all,       "xmap_enable_all",      NULL, XMAP_UINT32},
};

static char *xmap_trace_xia_cmd[7] = {	"adc_trace","baseline_history","trigger_filter","baseline_filter",
		"energy_filter","baseline_samples","energy_samples" };

static ELLLIST* drvList = NULL;
static ENV_PARAM XIAHOME;

xmap_command_t xmap_get_cmd( asynUser* pasynUser )
{
	epicsInt32 reason = pasynUser->reason;
	return xmap_commands[reason - XMAP_FIRST_CMD];
}

/*********************************************************/
/*                                                      */
/*  Configuration function to be called on the IOC     */
/*  shell at startup                                  */
/*                                                   */
/****************************************************/
int xmapConfig(char * portName, int nModule, char* initfile)
{
	int xiastatus = 0, i, dummy;
	asynStatus status = asynSuccess;
	xmap_driver_t *pDrv;
	xmap_channel_t *tmpchan;
	epicsInt32 *mcaBufPtr = NULL;
	char * xia_home_path;
	epicsTimeStamp startTime;
	char *strStartTime, *logFileName;

	printf("xmapConfig: asynport: \"%s\"\n"\
			"            Module number: %i\n"\
			"            Init file: \"%s\"\n", portName, nModule, initfile);

	/* allocate memory for local driver structure */
	pDrv = callocMustSucceed(1, sizeof(xmap_driver_t), "xmapConfig");

	pDrv->portName = epicsStrDup( portName );                      /* store the asyn port name */
	pDrv->initFile = calloc(40+1, sizeof(char));                   /* allocate 40 chars for the init filename (epics string record size) */
	strncpy(pDrv->initFile, initfile, 40);
	pDrv->localpasynUser = pasynManager->createAsynUser(0, 0);     /* Create one asyn user to be used for debugging at startup */
	pasynTrace->setTraceMask( pDrv->localpasynUser, 0x11);         /* strictly speaking not correct use of asyn trace masks
                                                                      however, it's the only way we can use asynPrint in the
                                                                      configuration function... */
	pDrv->start_mode = XMAP_START_ALL;
	pDrv->current_xiabuf = 'a';
	pDrv->resume = TRUE;
	pDrv->moduleNo = nModule;

	/* Default ADC settings */
	pDrv->trace_sampletime = 5000.0;
	pDrv->trace_large_buf = NULL;
	pDrv->trace_len = 0;
	pDrv->trace_mode = XMAP_TRACE_ADC;

	/* Allocating memory and assigning default values
	 * to the individual channel settings. */
	pDrv->chan = calloc(XMAP_NCHANS_MODULE, sizeof(xmap_channel_t));
	/* We allocate enough memory to hold 4 maximum sized MCA spectrums.
	 * after another and each of those will be pointed to by mcaChData... */
	mcaBufPtr = calloc(XMAP_NCHANS_MODULE * XMAP_MAX_MCA_BINS, sizeof(epicsInt32));
	for (i = 0; i < XMAP_NCHANS_MODULE; i++)
	{
		tmpchan = &(pDrv->chan[i]);
		tmpchan->nBins = XMAP_MAX_MCA_BINS;
		tmpchan->acquiring = 0;
		tmpchan->enabled = 1;

		/* Set up the mapping between address and xmap channel number:
		 * Modules, addresses and xmap channels all start from 0.
		 * So the xmap channel numbers can be mapped by: addr + 4 * module no. */
		pDrv->xch[i] = i + (nModule * XMAP_NCHANS_MODULE);

		/* When reading out the data we will get 4 spectrums one
		 * after another - so we set up 4 pointers into the big array... */
		tmpchan->mcaModuleData = mcaBufPtr;
		tmpchan->mcaChData = mcaBufPtr + (i * XMAP_MAX_MCA_BINS);
		printf("            modCh = %d globalCh = %d ptrmca = %p\n", i, pDrv->xch[i], tmpchan->mcaChData);
	}

	/* Link with higher level routines */
	pDrv->common.interfaceType = asynCommonType;
	pDrv->common.pinterface  = (void *)&xmap_Common;
	pDrv->common.drvPvt = pDrv;
	pDrv->int32.interfaceType = asynInt32Type;
	pDrv->int32.pinterface  = (void *)&xmap_Int32;
	pDrv->int32.drvPvt = pDrv;
	pDrv->uint32Digital.interfaceType = asynUInt32DigitalType;
	pDrv->uint32Digital.pinterface = (void *)&xmap_UInt32Digital;
	pDrv->uint32Digital.drvPvt = pDrv;
	pDrv->float64.interfaceType = asynFloat64Type;
	pDrv->float64.pinterface  = (void *)&xmap_Float64;
	pDrv->float64.drvPvt = pDrv;
	pDrv->int32Array.interfaceType = asynInt32ArrayType;
	pDrv->int32Array.pinterface  = (void *)&xmap_Int32Array;
	pDrv->int32Array.drvPvt = pDrv;
	pDrv->drvUser.interfaceType = asynDrvUserType;
	pDrv->drvUser.pinterface = (void *)&xmap_DrvUser;
	pDrv->drvUser.drvPvt = pDrv;

	/* registerport */
	status = pasynManager->registerPort(pDrv->portName,
			ASYN_MULTIDEVICE,
			1, /* autoconnect */
			0, /* medium priority */
			0); /* default stacksize */

	asynPrint( 	pDrv->localpasynUser, ASYN_TRACE_FLOW,
			"Created asynUser for port: %s\n", pDrv->portName);

	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig: ERROR: Can't register myself (port: %s)\n", pDrv->portName);
		return -1;
	}

	/* register the interfaces with asyn */
	status = pasynManager->registerInterface(pDrv->portName, &pDrv->common);
	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig :  ERROR:Can't register common. (port: %s)\n",pDrv->portName);
		return -1;
	}

	/* initializes */
	status = pasynInt32Base->initialize(pDrv->portName, &pDrv->int32);
	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig:  ERROR:Can't register int32. (port: %s)\n",pDrv->portName);
		return -1;
	}
	status = pasynFloat64Base->initialize(pDrv->portName, &pDrv->float64);
	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig: Can't register float64.\n");
		return -1;
	}
	status = pasynInt32ArrayBase->initialize(pDrv->portName, &pDrv->int32Array);
	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig: ERROR: Can't register int32Array. (port: %s)\n",pDrv->portName);
		return -1;
	}
	status = pasynUInt32DigitalBase->initialize(pDrv->portName, &pDrv->uint32Digital);
	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig: ERROR: Can't register uint32Digital.\n");
		return -1;
	}

	/* register interrupt source here  */
	pasynManager->registerInterruptSource(	pDrv->portName, &pDrv->int32, &pDrv->int32InterruptPvt);
	pasynManager->registerInterruptSource(	pDrv->portName, &pDrv->float64, &pDrv->float64InterruptPvt);
	pasynManager->registerInterruptSource(	pDrv->portName, &pDrv->int32Array, &pDrv->int32ArrayInterruptPvt);
	pasynManager->registerInterruptSource(	pDrv->portName, &pDrv->uint32Digital, &pDrv->uint32DigitalInterruptPvt);

	/* Register pasynManager interface */
	status = pasynManager->registerInterface(pDrv->portName, &pDrv->drvUser);
	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig ERROR: Can't register drvUser\n");
		return -1;
	}

	/* Connect pasynUser to device for debugging */
	status = pasynManager->connectDevice(pDrv->localpasynUser, pDrv->portName, 0);
	if (status != asynSuccess)
	{
		asynPrint( pDrv->localpasynUser, ASYN_TRACE_ERROR, "### xmapConfig, ERROR: connectDevice failed for scaler-sim\n");
		return -1;
	}

	/* if it is the first invocation of the config function,
	 * we initialize the xmap/handel library
	 * - and create the linked list of driver structures. */
	if (drvList == NULL)
	{
		printf("            Creating driver structure linked list\n");
		drvList = calloc(1, sizeof(ELLLIST));
		ellInit(drvList);
		ellAdd(drvList, (ELLNODE*)pDrv);

		printf("            Initializing Handel with \"%s\"\n", pDrv->initFile);
		xiastatus = xiaInit(pDrv->initFile);
		status = xmap_xia_checkError( pDrv->localpasynUser, xiastatus, "xiaInit()");
		if (status == asynError) return status;

		/* Register the epics exit function to be called when the IOC exits... */
		epicsAtExit((void*)xmapClose, NULL);

		xiastatus = xiaSetLogLevel(MD_ERROR);
		status = xmap_xia_checkError( pDrv->localpasynUser, xiastatus, "xiaSetLogLevel()");
		if (status == asynError) return status;

		/* Get the XIAHOME environment variable to define where to write the
		 * error log files to.		 */
		XIAHOME.name = epicsStrDup("XIAHOME");
		XIAHOME.pdflt = (char*)calloc(256, sizeof(char));
		xia_home_path = (char*)calloc(256, sizeof(char));
		envGetConfigParam(&XIAHOME, 256, xia_home_path);
		epicsTimeGetCurrent(&startTime);
		strStartTime = (char*)calloc(80, sizeof(char));
		epicsTimeToStrftime ( strStartTime, 80, "%Y-%m-%d_%H-%M-%S", &startTime );

		logFileName = (char*)calloc(256, sizeof(char));
		sprintf(logFileName, "%s\\epics-xmap_%s.log", xia_home_path, strStartTime);
		printf("            XIA log file: %s\n", logFileName);

		xiastatus = xiaSetLogOutput(logFileName);
		status = xmap_xia_checkError( pDrv->localpasynUser, xiastatus, "xiaSetLogOutput()");
		if (status == asynError) return status;

		printf("            Starting XIA system.\n");
		xiastatus = xiaStartSystem();
		if (status == asynError) return status;
	} else
	{
		/* just add the node to the existing list */
		ellAdd(drvList, (ELLNODE*)pDrv);
	}

	/* initialise the number of MCA bins on the hardware to its maximum */
/*	for (i=0; i < XMAP_NCHANS_MODULE; i++)
	{
		xch = pDrv->xch[i];
        dblNbins = (double)pDrv->chan[i].nBins;
        printf("            Setting ch: %d nbins: %d\n", xch, pDrv->chan[i].nBins);
		xiastatus = xiaSetAcquisitionValues(xch, "number_mca_channels", (void*)&dblNbins);
		status = xmap_xia_checkError(pDrv->localpasynUser, xiastatus, "number_mca_channels");
		if (status == asynError) return status;
	}
	*/
	/* and finally apply the operation to the board */
	printf("            Finally applying board operation\n");
	xiastatus = xiaBoardOperation(pDrv->xch[0], "apply", (void*)&dummy);
	status = xmap_xia_checkError(pDrv->localpasynUser, xiastatus, "apply");
	if (status == asynError) return status;

	printf("xmapConfig: Done!\n\n");
	return 0;
}

void xmapClose(void)
{
	int status;
	printf("Closing down XMAP port");
	status = xiaExit();
	if (status == XIA_SUCCESS)
	{
		printf("### XIA XMAP closed down successfully. You can now exit the iocshell\n");
		return;
	}
	printf("xiaExit() error: %d\n", status);
	return;
}

/* Figure out the type of support call a given command is for:
   MCA: asyn command from an MCA record.
   XIA: asyn command that translate directly into a XIA/Handel command.
   XMAP: an asyn command that is local to this driver - i.e. a combination
         of XIA/MCA commands or local state variables. */
xmap_support_t xmap_get_support_type(epicsInt32 reason)
{
	if ( reason < 0 || reason > XMAP_TOTAL_CMD)
		return XMAP_CMD_INVALID;

	if ( reason >= XMAP_FIRST_MCA_CMD && reason <= XMAP_LAST_MCA_CMD)
		return XMAP_MCA;
	if ( reason >= XMAP_FIRST_XIA_CMD && reason <= XMAP_LAST_XIA_CMD)
	{
		if (reason >= xia_peaking_time && reason <= xia_decay_time)
			return XMAP_XIA_ACQ;
		if (reason >= xia_apply && reason <= xia_mapping_pixel_next)
			return XMAP_XIA_BOARD;
		if (reason >= xia_mca_length && reason <= xia_buffer_overrun)
			return XMAP_XIA_RUN;
	}
	if ( reason >= XMAP_FIRST_CMD && reason <= XMAP_LAST_CMD)
	{
		if (reason >= xmap_mca_collect && reason < xmap_some_command)
		{
			/* If the command is part of the extended set of MCA commands
			 * - extensions to what the MCA module/record provides.	 */
			return XMAP_MCA;
		}
		return XMAP_XMAP;
	}

	/* if everything fails... */
	return XMAP_CMD_INVALID;
}


/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
static void xmap_Report(void *drvPvt, FILE *fp, int details)
{
	/* Implement some reporting routine here, printing out various info bits
	   about the state of the hardware and driver... */
}

static asynStatus xmap_Connect(void *drvPvt, asynUser *pasynUser)
{
	int signal;
	pasynManager->getAddr(pasynUser, &signal);
	pasynManager->exceptionConnect(pasynUser);
	return(asynSuccess);
}

static asynStatus xmap_Disconnect(void *drvPvt, asynUser *pasynUser)
{
	pasynManager->exceptionDisconnect(pasynUser);
	return(asynSuccess);
}

static asynStatus xmap_getBounds(void *drvPvt, asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high)
{
	*low = 0;
	*high = 0;
	/* not usually implemented... See asyn documentation for details if needed. */
	return( asynSuccess );
}


/*****         asynInt32 methods       ******/
static asynStatus xmap_int32Write(void *drvPvt, asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	xmap_support_t suppType;
	epicsInt32 reason = pasynUser->reason;
	xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;

	suppType = xmap_get_support_type(reason);
	switch( suppType )
	{
	case XMAP_MCA:
		status = xmap_mca_int_write (pPvt, pasynUser, value);
		break;
	case XMAP_XMAP:
		break;
	case XMAP_CMD_INVALID:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_int32Write Invalid reason (%d)!\n", reason);
		status = asynError;
		break;
	default:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_int32Write This type of command 0x%2.2X is not supported by float64 write!\n",
				suppType);
		status = asynError;
		break;
	}
	return( status );
}

static asynStatus xmap_int32Read(void *drvPvt, asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 dummy;
	xmap_support_t suppType;
	epicsInt32 reason = pasynUser->reason;
	xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;

	suppType = xmap_get_support_type(reason);
	switch( suppType )
	{
	case XMAP_XIA_RUN:
		status = xmap_xia_run_val(pasynUser, value, &dummy, XMAP_INT32);
		break;
	case XMAP_MCA:
		status = xmap_mca_int_read(pPvt, pasynUser, value);
		break;
	case XMAP_XMAP:
		break;
	case XMAP_CMD_INVALID:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_int32Read Invalid reason (%d)!\n", reason);
		status = asynError;
		break;
	default:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_int32Read This type of command 0x%2.2X is not supported by float64 write!\n",
				suppType);
		status = asynError;
		break;
	}
	return( status );
}


/****          Asyn int32 digital routines      *******/
static asynStatus xmap_uint32DigitalWrite(void *drvPvt, asynUser *pasynUser, epicsUInt32 value, epicsUInt32 mask)
{
	asynStatus status = asynSuccess;
	xmap_support_t suppType;
	epicsInt32 reason = pasynUser->reason;
	xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;

	suppType = xmap_get_support_type(reason);
	switch( suppType )
	{
	case XMAP_XIA_BOARD:
		status = xmap_xia_board_op(pasynUser, &(pPvt->current_xiabuf));
		break;
	case XMAP_MCA:
		status = xmap_mca_int_write (pPvt, pasynUser, (epicsInt32)value);
		break;
	case XMAP_XMAP:
		if (reason == xmap_trace)
		{
			if (value >= 0 && value <= XMAP_TRACE_ENERGY_SAMPLES)
				pPvt->trace_mode = value;
			else status = asynError;
			break;
		}
		break;
	case XMAP_CMD_INVALID:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_uint32DigitalWrite Invalid reason (%d)!\n", reason);
		status = asynError;
		break;
	default:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_uint32DigitalWrite This type of command 0x%2.2X is not supported by float64 write!\n",
				suppType);
		status = asynError;
		break;
	}
	return( status );
}

static asynStatus xmap_uint32DigitalRead( void *drvPvt, asynUser *pasynUser, epicsUInt32 *value, epicsUInt32 mask)
{
	asynStatus status = asynSuccess;
	epicsInt32 dummy;
	xmap_support_t suppType;
	epicsInt32 reason = pasynUser->reason;
	xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;

	suppType = xmap_get_support_type(reason);
	switch( suppType )
	{
	case XMAP_XIA_RUN:
		status = xmap_xia_run_val(pasynUser, value, &dummy, XMAP_UINT32);
		break;
	case XMAP_MCA:
		status = xmap_mca_int_read(pPvt, pasynUser, value);
		break;
	case XMAP_XMAP:
		break;
	case XMAP_CMD_INVALID:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_uint32DigitalRead Invalid reason (%d)!\n", reason);
		status = asynError;
		break;
	default:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_uint32DigitalRead This type of command 0x%2.2X is not supported by float64 write!\n",
				suppType);
		status = asynError;
		break;
	}
	return( status );
}

/* asynFloat64 methods */
static asynStatus xmap_float64Write(void *drvPvt, asynUser *pasynUser, epicsFloat64 value)
{
	asynStatus status = asynSuccess;
	xmap_support_t suppType;
	epicsInt32 reason = pasynUser->reason;
	xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;

	suppType = xmap_get_support_type(reason);
	switch( suppType )
	{
	case XMAP_XIA_ACQ:
		status = xmap_xia_acq_val( pasynUser, &value, XMAP_SET);
		break;
	case XMAP_MCA:
		status = xmap_mca_float_write (pPvt, pasynUser, value);
		break;
	case XMAP_XMAP:
		if (reason == xmap_trace)
			pPvt->trace_sampletime = value;
		break;
	case XMAP_CMD_INVALID:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_float64Write Invalid reason (%d)!\n", reason);
		status = asynError;
		break;
	default:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_float64Write This type of command 0x%2.2X is not supported by float64 write!\n",
				suppType);
		status = asynError;
		break;
	}
	return( status );
}

static asynStatus xmap_float64Read(void *drvPvt, asynUser *pasynUser, epicsFloat64 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 dummy;
	epicsInt32 reason = pasynUser->reason;
	xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;
	xmap_support_t suppType;

	suppType = xmap_get_support_type(reason);
	switch( suppType )
	{
	case XMAP_XIA_ACQ:
		status = xmap_xia_acq_val( pasynUser, value, XMAP_GET);
		break;
	case XMAP_XIA_RUN:
		status = xmap_xia_run_val(pasynUser, value, &dummy, XMAP_FLOAT64);
		break;
	case XMAP_MCA:
		status = xmap_mca_float_read (pPvt, pasynUser, value);
		break;
	case XMAP_XMAP:
		if (reason == xmap_trace)
			*value = pPvt->trace_sampletime;
		break;
	case XMAP_CMD_INVALID:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_float64Read Invalid reason (%d)!\n", reason);
		status = asynError;
		break;
	default:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_float64Read This type of command 0x%2.2X is not supported by float64 read!\n",
				suppType);
		status = asynError;
		break;
	}
	return( status );
}

/* asynInt32Array methods */
static asynStatus xmap_int32ArrayRead(void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t maxChans, size_t *nactual)
{
	asynStatus status = asynSuccess;
	int xiastatus;
	epicsInt32 channel;
	epicsUInt32 nord;
	xmap_support_t suppType;
	xmap_command_t cmd;

	epicsInt32 reason = pasynUser->reason;
	xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;
	pasynManager->getAddr(pasynUser, &channel);		/* Get the channel number */

	suppType = xmap_get_support_type(reason);
	switch( suppType )
	{
	case XMAP_XIA_RUN:
		break;
	case XMAP_MCA:
		status = xmap_mca_int_array_read (pPvt, pasynUser, data, maxChans, nactual);
		break;
	case XMAP_XMAP:
		cmd = xmap_get_cmd( pasynUser );
		if (reason == xia_mca)
		{
			xiastatus = xiaGetRunData( channel, "mca_length", (void*)&nord);
			status = xmap_xia_checkError( pasynUser, xiastatus, "mca_length" );
			asynPrint(pasynUser, ASYN_TRACE_FLOW, "drvXmap::xmap_int32ArrayRead (mca): number of bins read: %d\n", nord);
			if (nord > maxChans)
			{
				asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_int32ArrayRead (mca): too many bins: %d Max allowed: %d\n",
						nord, maxChans);
				status = asynError;
			} else *nactual = nord;
			break;
		} else if (reason == xmap_trace)
		{
			status = xmap_get_trace( pasynUser, pPvt, data, (epicsInt32)maxChans, (epicsInt32*)nactual);
		}
		break;
	case XMAP_CMD_INVALID:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_int32ArrayRead Invalid reason (%d)!\n", reason);
		status = asynError;
		break;
	default:
		asynPrint(	pasynUser, ASYN_TRACE_ERROR, "drvXmap::xmap_int32ArrayRead This type of command 0x%2.2X is not supported by float64 read!\n",
				suppType);
		status = asynError;
		break;
	}
	return( status );
}

static asynStatus xmap_int32ArrayWrite(void *drvPvt, asynUser *pasynUser, epicsInt32 *data, size_t maxChans)
{
	asynStatus status = asynSuccess;
	return( status );
}

/* Get one shot of ADC trace from the given channel. This is also known as
   'oscilloscope mode'. */
asynStatus xmap_get_trace(	asynUser* pasynUser, xmap_driver_t *pPvt,
		epicsInt32* data, epicsInt32 maxlen, epicsInt32* actuallen)
{
	asynStatus status = asynSuccess;
	int xiastatus, channel;
	double info[2] = {0.0, (double)pPvt->trace_sampletime};
	epicsInt32* buf = data;

	pasynManager->getAddr(pasynUser, &channel);
	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
			"drvXmap::xmap_get_trace: Getting trace on ch%d (%d %.2f)\n",channel, maxlen, pPvt->trace_sampletime);

	xiastatus = xiaDoSpecialRun( channel, xmap_trace_xia_cmd[pPvt->trace_mode], (void*)info);
	status = xmap_xia_checkError( pasynUser, xiastatus, xmap_trace_xia_cmd[pPvt->trace_mode] );
	if (status == asynError) return asynError;

	/* Only ask for length if we haven't done so before. The length of the ADC
	   trace data from the xmap is constant according to the manual so we only read it once! */
	if (pPvt->trace_len == 0)
	{
		asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvXmap::xmap_get_trace: Getting trace_length\n");

		xiastatus = xiaGetSpecialRunData( channel, "adc_trace_length", (void*) &(pPvt->trace_len));
		status = xmap_xia_checkError( pasynUser, xiastatus, "adc_trace_length" );
		if (status == asynError) return asynError;

		asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvXmap::xmap_get_trace: trace_length = %d\n", pPvt->trace_len);
	}

	/* If the lenght of the trace is bigger than NELM (maxlen),
	   then we use a local (bigger) buffer to read the data into */
	if (pPvt->trace_len > maxlen)
	{
		asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvXmap::xmap_get_trace: Length of data (%d) is greater than allocated memory (%d)\n",
				pPvt->trace_len, maxlen);
		/* if the local big buffer has not been allocated yet we do so now... */
		if (pPvt->trace_large_buf == NULL)
		{
			pPvt->trace_large_buf = calloc( pPvt->trace_len, sizeof(epicsInt32));
			if (pPvt->trace_large_buf == NULL) return asynError;
		}
		/* set the readout buffer to be the local big one */
		buf = pPvt->trace_large_buf;
	}
	else *actuallen = pPvt->trace_len;

	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
			"drvXmap::xmap_get_trace: reading out trace\n");

	xiastatus = xiaGetSpecialRunData( channel, "adc_trace", (void*) buf);
	status = xmap_xia_checkError( pasynUser, xiastatus, "adc_trace" );
	if (status == asynError) return status;

	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
			"drvXmap::xmap_get_trace: trace first samples: %d %d %d %d %d\n",
			buf[0], buf[1], buf[2], buf[3], buf[4]);

	/* If the data was read into the larger local buffer, then we
	   just copy it over into the smaller buffer of the waveform rec. */
	if (buf == pPvt->trace_large_buf)
		memcpy( data, buf, pPvt->trace_len * sizeof(epicsInt32));

	return status;
}




















/* asynDrvUser methods */
static asynStatus xmap_drvUserCreate(void *drvPvt, asynUser *pasynUser, const char *drvInfo, const char **pptypeName, size_t *psize)
{
	epicsInt32 i;
	char *pstring;
	/*xmap_driver_t *pPvt = (xmap_driver_t*)drvPvt;*/
	xmap_command_t xiaCmd;
	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
			"drvXmap::drvUserCreate attempting to create cmd: %s (XMAP_TOTAL_CMD: %d)\n", drvInfo, XMAP_TOTAL_CMD);

	/*	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvXmap::drvUserCreate\tchecking XMAP local commands...(%d - %d -1= %d\n",
				XMAP_LAST_CMD, XMAP_FIRST_CMD,(XMAP_LAST_CMD - XMAP_FIRST_CMD-1));*/
	for (i = 0; i < XMAP_N_CMDS; i++)
	{
		pstring = xmap_commands[i].commandString;
		if (epicsStrCaseCmp( drvInfo, pstring) == 0)
		{
			pasynUser->reason = xmap_commands[i].command;
			if (pptypeName) *pptypeName = epicsStrDup(pstring);
			if (psize) *psize = sizeof( xmap_commands[i].command);
			asynPrint(	pasynUser, ASYN_TRACE_FLOW,
					"drvXmap::drvUserCreate, xmap command created: %s\n", pstring);
			return( asynSuccess );
		}
	}

	/*	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvXmap::drvUserCreate\tchecking MCA commands... (%d - %d -1 = %d\n",
				XMAP_LAST_MCA_CMD, XMAP_FIRST_MCA_CMD,(XMAP_LAST_MCA_CMD - XMAP_FIRST_MCA_CMD -1));*/
	for (i = 0; i < MAX_MCA_COMMANDS; i++)
	{
		pstring = mcaCommands[i].commandString;
		if (epicsStrCaseCmp( drvInfo, pstring) == 0)
		{
			pasynUser->reason = mcaCommands[i].command;
			if (pptypeName) *pptypeName = epicsStrDup(pstring);
			if (psize) *psize = sizeof(mcaCommands[i].command);
			asynPrint(	pasynUser, ASYN_TRACE_FLOW,
					"drvXmap::drvUserCreate, MCA command created: %s\n", pstring);
			return( asynSuccess );
		}
	}

	/*	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"drvXmap::drvUserCreate\tchecking XIA/Handel commands... (%d - %d -1 = %d\n",
				XMAP_LAST_XIA_CMD, XMAP_FIRST_XIA_CMD,(XMAP_LAST_XIA_CMD - XMAP_FIRST_XIA_CMD -1));*/
	for (i = 0; i < XMAP_XIA_CMDS; i++)
	{
		xiaCmd = xmap_xia_get_cmdreason(i + XMAP_FIRST_XIA_CMD);
		pstring = xiaCmd.commandString;
		if (epicsStrCaseCmp( drvInfo, pstring) == 0)
		{
			pasynUser->reason = xiaCmd.command;
			if (pptypeName)	*pptypeName = epicsStrDup(pstring);
			if (psize) *psize = sizeof(xiaCmd.command);
			asynPrint(	pasynUser, ASYN_TRACE_FLOW,
					"drvXmap::drvUserCreate, XIA/Handel command created: %s\n", pstring);
			return( asynSuccess );
		}
	}

	asynPrint(	pasynUser, ASYN_TRACE_ERROR,
			"drvXmap::drvUserCreate, unknown command: %s\n", drvInfo);
	return( asynError );
}

static asynStatus xmap_drvUserGetType(void *drvPvt, asynUser *pasynUser, const char **pptypeName, size_t *psize)
{
	int reason = pasynUser->reason;
	xmap_support_t suppType;
	char * pString;
	suppType = xmap_get_support_type(reason);

	if (suppType & XMAP_CMD_INVALID) return asynError;

	if (suppType & (XMAP_XIA_ACQ | XMAP_XIA_RUN | XMAP_XIA_BOARD))
	{
		pString = xmap_xia_get_cmdreason( reason ).commandString;
	}
	else if (suppType & XMAP_MCA)
	{
		pString = mcaCommands[reason].commandString;
	}
	else if (suppType & XMAP_XMAP)
	{
		pString = xmap_commands[reason - XMAP_FIRST_CMD].commandString;
	} else return asynError;

	*psize = 0;
	if ( pptypeName )
		*pptypeName = epicsStrDup( pString );
	if ( psize )
		*psize = sizeof( reason );
	return( asynSuccess );
}

static asynStatus xmap_drvUserDestroy(void *drvPvt, asynUser *pasynUser)
{
	/*	asynPrint(	pasynUser, ASYN_TRACE_FLOW,
				"devMr1394Asyn::drvUserDestroy. Function not implemented\n");*/
	return( asynSuccess );
}
